@@ -6,7 +6,7 @@
 ##
 ## All Rights Reserved
 ##
-## See ll/__init__.py for the license
+## See ll/xist/__init__.py for the license
 
 
 """
@@ -31,6 +31,8 @@ this::
 	from ll import make
 
 	class MyProject(make.Project):
+		name = "Acme.MyProject"
+
 		def create(self):
 			make.Project.create(self)
 			source = self.add(make.FileAction("foo.txt"))
@@ -404,8 +406,7 @@ class Action(object):
 		"""
 		for input in self:
 			yield input
-			for subinput in input.iterallinputs():
-				yield subinput
+			yield from input.iterallinputs()
 
 	def findpaths(self, input):
 		"""
@@ -478,10 +479,8 @@ class CollectAction(TransformAction):
 		return self
 
 	def __iter__(self):
-		for input in TransformAction.__iter__(self):
-			yield input
-		for input in self.otherinputs:
-			yield input
+		yield from TransformAction.__iter__(self)
+		yield from self.otherinputs
 
 	@report
 	def get(self, project, since):
@@ -748,8 +747,7 @@ class GetAttrAction(TransformAction):
 		self.attrname = attrname
 
 	def __iter__(self):
-		for input in TransformAction.__iter__(self):
-			yield input
+		yield from TransformAction.__iter__(self)
 		yield self.attrname
 
 	def getkwargs(self):
@@ -774,10 +772,8 @@ class CallAction(Action):
 
 	def __iter__(self):
 		yield self.func
-		for input in self.args:
-			yield input
-		for input in self.kwargs.values():
-			yield input
+		yield from self.args
+		yield from self.kwargs.values()
 
 	def getargs(self):
 		return (self.func,) + self.args
@@ -820,10 +816,8 @@ class CallAttrAction(Action):
 	def __iter__(self):
 		yield self.obj
 		yield self.attrname
-		for input in self.args:
-			yield input
-		for input in self.kwargs.values():
-			yield input
+		yield from self.args
+		yield from self.kwargs.values()
 
 	def getargs(self):
 		return (self.obj, self.attrname) + self.args
@@ -873,8 +867,7 @@ class ModeAction(TransformAction):
 		self.mode = mode
 
 	def __iter__(self):
-		for input in TransformAction.__iter__(self):
-			yield input
+		yield from TransformAction.__iter__(self)
 		yield self.mode
 
 	def getkwargs(self):
@@ -908,8 +901,7 @@ class OwnerAction(TransformAction):
 		self.group = group
 
 	def __iter__(self):
-		for input in TransformAction.__iter__(self):
-			yield input
+		yield from TransformAction.__iter__(self)
 		yield self.user
 		yield self.group
 
@@ -964,10 +956,8 @@ class ModuleAction(TransformAction):
 		return self
 
 	def __iter__(self):
-		for input in TransformAction.__iter__(self):
-			yield input
-		for input in self.inputs:
-			yield input
+		yield from TransformAction.__iter__(self)
+		yield from self.inputs
 
 	def execute(self, project, data):
 		key = self.getkey()
@@ -1115,6 +1105,10 @@ class Project(dict):
 	is responsible for initiating the build process and for generating a report
 	about the progress of the build process.
 	"""
+
+	# Will be used in log messages and notifications
+	name = "ll.make"
+
 	def __init__(self):
 		super().__init__()
 		self.actionscalled = 0
@@ -1143,7 +1137,7 @@ class Project(dict):
 		self.showtime = self._getenvbool("LL_MAKE_SHOWTIME", True)
 		self.showtimestamps = self._getenvbool("LL_MAKE_SHOWTIMESTAMPS", False)
 		self.showdata = self._getenvbool("LL_MAKE_SHOWDATA", False)
-		self.growl = self._getenvbool("LL_MAKE_GROWL", False)
+		self.notify = self._getenvbool("LL_MAKE_NOTIFY", False)
 
 	def __repr__(self):
 		return "<{}.{} with {} targets at {:#x}>".format(self.__module__, self.__class__.__name__, len(self), id(self))
@@ -1480,7 +1474,7 @@ class Project(dict):
 		p.add_argument("targets", metavar="target", help="Target to be built", nargs="*")
 		p.add_argument("-x", "--ignoreerrors", dest="ignoreerrors", help="Ignore errors? (default: %(default)s)", action=misc.FlagAction, default=self.ignoreerrors)
 		p.add_argument("-c", "--color", dest="color", help="Use colored output? (default: %(default)s)", action=misc.FlagAction, default=self.color)
-		p.add_argument("-g", "--growl", dest="growl", help="Issue growl notification after the build? (default: %(default)s)", action=misc.FlagAction, default=self.growl)
+		p.add_argument("-y", "--notify", dest="notify", help="Issue notification after the build? (default: %(default)s)", action=misc.FlagAction, default=self.notify)
 		p.add_argument("-a", "--showaction", dest="showaction", help="Show actions? (default: %(default)s)", choices=actions, default=action2name(self.showaction))
 		p.add_argument("-s", "--showstep", dest="showstep", help="Show steps? (default: %(default)s)", choices=actions, default=action2name(self.showstep))
 		p.add_argument("-n", "--shownote", dest="shownote", help="Show notes? (default: %(default)s)", choices=actions, default=action2name(self.shownote))
@@ -1499,7 +1493,7 @@ class Project(dict):
 		args = p.parse_args(args)
 		self.ignoreerrors = args.ignoreerrors
 		self.color = args.color
-		self.growl = args.growl
+		self.notify = args.notify
 		self.showaction = args.showaction
 		self.showstep = args.showstep
 		self.shownote = args.shownote
@@ -1559,57 +1553,43 @@ class Project(dict):
 
 			self.buildno += 1 # increment build number so that actions that stored the old one can detect a new build round
 
-			for target in targets:
-				data = self._get(target, bigcrunch)
-			now = datetime.datetime.utcnow()
+			success = False
+			try:
+				if self.notify:
+					self.notifystart()
+				for target in targets:
+					data = self._get(target, bigcrunch)
 
-			if self.showsummary:
-				args = []
-				self.write(
-					"built ",
-					s4action(self.__class__.__module__, ".", self.__class__.__name__),
-					": ",
-					s4data(format(len(self))),
-					" registered targets; ",
-					s4data(format(self.actionscalled)),
-					" actions called; ",
-					s4data(format(self.stepsexecuted)),
-					" steps executed; ",
-					s4data(format(self.filesread)),
-					" files/",
-					s4data(format(self.bytesread)),
-					" bytes read; ",
-					s4data(format(self.fileswritten)),
-					" files/",
-					s4data(format(self.byteswritten)),
-					" bytes written; ",
-					s4data(format(self.actionsfailed)),
-					" actions failed",
-				)
-				if self.showtime:
-					self.write(" [t+", self.strtimedelta(now-self.starttime), "]")
-				self.writeln()
-			if self.growl:
-				try:
-					module = sys.modules[self.__class__.__module__]
-				except KeyError:
-					filename = "???.py"
-				else:
-					filename = os.path.abspath(module.__file__)
-				cmd = "growlnotify -i py -n ll.make 'll.make done in {}'".format(self.strtimedelta(now-self.starttime))
-				import subprocess
-				pipe = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE).stdin
-				def write(s):
-					pipe.write(s.encode("utf-8"))
-				write("{}\n".format(filename))
-				write("{}.{}\n".format(self.__class__.__module__, self.__class__.__name__))
-				write("{:,} registered targets, ".format(len(self)))
-				write("{:,} actions called, ".format(self.actionscalled))
-				write("{:,} steps executed, ".format(self.stepsexecuted))
-				write("{:,} files/{:,} bytes read, ".format(self.filesread, self.bytesread))
-				write("{:,} files/{:,} bytes written, ".format(self.fileswritten, self.byteswritten))
-				write("{:,} actions failed".format(self.actionsfailed))
-				pipe.close()
+				if self.showsummary:
+					args = []
+					self.write(
+						"built project ",
+						s4action(self.name),
+						": ",
+						s4data(format(len(self))),
+						" registered targets; ",
+						s4data(format(self.actionscalled)),
+						" actions called; ",
+						s4data(format(self.stepsexecuted)),
+						" steps executed; ",
+						s4data(format(self.filesread)),
+						" files/",
+						s4data(format(self.bytesread)),
+						" bytes read; ",
+						s4data(format(self.fileswritten)),
+						" files/",
+						s4data(format(self.byteswritten)),
+						" bytes written; ",
+						s4data(format(self.actionsfailed)),
+						" actions failed",
+					)
+					if self.showtime:
+						self.write(" [t+", self.strtimedelta(datetime.datetime.utcnow()-self.starttime), "]")
+					self.writeln()
+				success = True
+			finally:
+				if self.notify:
+					self.notifyfinish(datetime.datetime.utcnow()-self.starttime, success)
 
 	def buildwithargs(self, args=None):
 		"""
@@ -1647,6 +1627,44 @@ class Project(dict):
 		"""
 		self.write(*texts)
 
+	def notifystart(self):
+		cmd = [
+			"/Applications/terminal-notifier.app/Contents/MacOS/terminal-notifier",
+			"-remove",
+			misc.sysinfo.scriptname,
+		]
+
+		import subprocess
+		with open("/dev/null", "wb") as f:
+			status = subprocess.call(cmd, stdout=f)
+
+	def notifyfinish(self, duration, success):
+		msgs = []
+		if self.stepsexecuted:
+			msgs.append("{:,} steps".format(self.stepsexecuted))
+		if self.fileswritten:
+			msgs.append("{:,} files".format(self.fileswritten))
+		if self.byteswritten:
+			msgs.append("{:,} bytes".format(self.byteswritten))
+		if not msgs:
+			msgs.append("nothing to do")
+
+		cmd = [
+			"/Applications/terminal-notifier.app/Contents/MacOS/terminal-notifier",
+			"-title",
+			self.name,
+			"-subtitle",
+			"{} {}".format("finished after" if success else "failed after", self.strtimedelta(duration)),
+			"-message",
+			"; ".join(msgs),
+			"-group",
+			misc.sysinfo.scriptname,
+		]
+
+		import subprocess
+		with open("/dev/null", "wb") as f:
+			status = subprocess.call(cmd, stdout=f)
+
 	def warn(self, warning, stacklevel):
 		"""
 		Issue a warning through the Python warnings framework
